/*
  Copyright 1998 David Rysdam
  This file is released under the GPL.  Details can be
  found in the file COPYING accompanying this distribution.
*/

#include<stdio.h>
#include<ctype.h>
#include <string.h>
#include <oledecod.h>
#include"xltypes.h"
#include <stdlib.h>

void print_sheet(void);
void free_sheet(void);
void print_value(char *value);
int getshort(char *rec,int offset);
char *format_double(char *rec,int offset);
char *gettypename(long rectype);
void printhexy(unsigned char rec[], long reclen);
void process_item (int rectype, int reclen, char *rec); 
void allocate(int row,int col);

int main(int argc, char *argv[])
{
  FILE *INPUT;
  long rectype;
  long reclen;
  unsigned char rec[MAX_MS_RECSIZE];
  int itemsread=1;
  char *rectypename;
  char *filename;
  int result;
  pps_entry *stream_tree;
  U32 root_stream;
  U32 stream;
  
  filename = argv[1];
  result = OLEdecode(filename,&stream_tree,&root_stream,0);
  if (result) {
     fprintf(stderr,"Error parsing OLE file %d\n",result);
     exit (result);
  }   

  for (stream = stream_tree[root_stream].dir;
       stream!=0xffffffff && strcmp(stream_tree[stream].name, "Book")
        && strcmp(stream_tree[stream].name,"Workbook");
       stream = stream_tree[stream].next);
       
   if (stream == 0xfffffff) {
      fprintf(stderr,"No excel workbook stream found\n");
      exit(1);
   }   
  INPUT = fopen(stream_tree[stream].filename, "r");
  if(INPUT == NULL) {
    fprintf(stderr,"An error occurred opening the file\n");
    exit(1);
  }  
  while(itemsread){
    rectype = 0;
    itemsread = fread(&rectype, 2, 1, INPUT);
    
    if(itemsread == 0)
      break;

    reclen=0;

    rectypename = gettypename(rectype);
    itemsread = fread(&reclen, 2, 1, INPUT);
    if(reclen){
      itemsread = fread(rec, 1, reclen, INPUT);
      rec[reclen] = '\0';
    }
    process_item(rectype,reclen,rec);
  }
  fclose(INPUT);
  freeOLEtree(stream_tree); 
  return 0;
}

void printhexy(unsigned char rec[], long reclen)
{
  char charoutput[17];
  int i;
  int poscount;

  poscount = (reclen - reclen%16) + 16;
  strcpy(charoutput, "                ");
  charoutput[16] = '\0';

  for(i=0;i<poscount;i++){
    if(i<reclen){
      printf("%02X ", rec[i]);
      charoutput[i%16] = isprint(rec[i])?rec[i]:'.';
    }
    else
      printf("   ");
    if((i+1)%16==0){
      printf(" %s\n", charoutput);
      strcpy(charoutput, "                ");
    }
  }
}

char *gettypename(long rectype)
{
  switch(rectype)
    {
    case MS1904:
      return "1904";
    case ADDIN:
      return "ADDIN"; 
    case ADDMENU:
      return "ADDMENU"; 
    case ARRAY:
      return "ARRAY"; 
    case AUTOFILTER:
      return "AUTOFILTER"; 
    case AUTOFILTERINFO:
      return "AUTOFILTERINFO"; 
    case BACKUP:
      return "BACKUP"; 
    case BLANK:
      return "BLANK"; 
    case BOF:
      return "BOF"; 
    case BOOKBOOL:
      return "BOOKBOOL"; 
    case BOOLERR:
      return "BOOLERR"; 
    case BOTTOMMARGIN:
      return "BOTTOMMARGIN"; 
    case BOUNDSHEET:
      return "BOUNDSHEET"; 
    case CALCCOUNT:
      return "CALCCOUNT"; 
    case CALCMODE:
      return "CALCMODE"; 
    case CODEPAGE:
      return "CODEPAGE"; 
    case COLINFO:
      return "COLINFO"; 
    case CONTINUE:
      return "CONTINUE"; 
    case COORDLIST:
      return "COORDLIST"; 
    case COUNTRY:
      return "COUNTRY"; 
    case CRN:
      return "CRN"; 
    case DBCELL:
      return "DBCELL"; 
    case DCON:
      return "DCON"; 
    case DCONNAME:
      return "DCONNAME"; 
    case DCONREF:
      return "DCONREF"; 
    case DEFAULTROWHEIGHT:
      return "DEFAULTROWHEIGHT"; 
    case DEFCOLWIDTH:
      return "DEFCOLWIDTH"; 
    case DELMENU:
      return "DELMENU"; 
    case DELTA:
      return "DELTA"; 
    case DIMENSIONS:
      return "DIMENSIONS"; 
    case DOCROUTE:
      return "DOCROUTE"; 
    case EDG:
      return "EDG"; 
    case MSEOF:
      return "EOF"; 
    case EXTERNCOUNT:
      return "EXTERNCOUNT"; 
    case EXTERNNAME:
      return "EXTERNNAME"; 
    case EXTERNSHEET:
      return "EXTERNSHEET"; 
    case FILEPASS:
      return "FILEPASS"; 
    case FILESHARING:
      return "FILESHARING"; 
    case FILESHARING2:
      return "FILESHARING2"; 
    case FILTERMODE:
      return "FILTERMODE"; 
    case FNGROUPCOUNT:
      return "FNGROUPCOUNT"; 
    case FNGROUPNAME:
      return "FNGROUPNAME"; 
    case FONT:
      return "FONT"; 
    case FOOTER:
      return "FOOTER"; 
    case FORMAT:
      return "FORMAT"; 
    case FORMULA:
      return "FORMULA"; 
    case GCW:
      return "GCW"; 
    case GRIDSET:
      return "GRIDSET"; 
    case GUTS:
      return "GUTS"; 
    case HCENTER:
      return "HCENTER"; 
    case HEADER:
      return "HEADER"; 
    case HIDEOBJ:
      return "HIDEOBJ"; 
    case HORIZONTALPAGEBREAKS:
      return "HORIZONTALPAGEBREAKS"; 
    case IMDATA:
      return "IMDATA"; 
    case INDEX:
      return "INDEX"; 
    case INTERFACEEND:
      return "INTERFACEEND"; 
    case INTERFACEHDR:
      return "INTERFACEHDR"; 
    case ITERATION:
      return "ITERATION"; 
    case LABEL:
      return "LABEL"; 
    case LEFTMARGIN:
      return "LEFTMARGIN"; 
    case LHNGRAPH:
      return "LHNGRAPH"; 
    case LHRECORD:
      return "LHRECORD"; 
    case LPR:
      return "LPR"; 
    case MMS:
      return "MMS"; 
    case MULBLANK:
      return "MULBLANK"; 
    case MULRK:
      return "MULRK"; 
    case NAME:
      return "NAME"; 
    case NOTE:
      return "NOTE"; 
    case NUMBER:
      return "NUMBER"; 
    case OBJ:
      return "OBJ"; 
    case OBJPROTECT:
      return "OBJPROTECT"; 
    case OBPROJ:
      return "OBPROJ"; 
    case OLESIZE:
      return "OLESIZE"; 
    case PALETTE:
      return "PALETTE"; 
    case PANE:
      return "PANE"; 
    case PASSWORD:
      return "PASSWORD"; 
    case PLS:
      return "PLS"; 
    case PRECISION:
      return "PRECISION"; 
    case PRINTGRIDLINES:
      return "PRINTGRIDLINES"; 
    case PRINTHEADERS:
      return "PRINTHEADERS"; 
    case PROTECT:
      return "PROTECT"; 
    case PUB:
      return "PUB"; 
    case RECIPNAME:
      return "RECIPNAME"; 
    case REFMODE:
      return "REFMODE"; 
    case RIGHTMARGIN:
      return "RIGHTMARGIN"; 
    case RK:
      return "RK"; 
    case ROW:
      return "ROW"; 
    case RSTRING:
      return "RSTRING"; 
    case SAVERECALC:
      return "SAVERECALC"; 
    case SCENARIO:
      return "SCENARIO"; 
    case SCENMAN:
      return "SCENMAN"; 
    case SCENPROTECT:
      return "SCENPROTECT"; 
    case SCL:
      return "SCL"; 
    case SELECTION:
      return "SELECTION"; 
    case SETUP:
      return "SETUP"; 
    case SHRFMLA:
      return "SHRFMLA"; 
    case SORT:
      return "SORT"; 
    case SOUND:
      return "SOUND"; 
    case STANDARDWIDTH:
      return "STANDARDWIDTH"; 
    case STRING:
      return "STRING"; 
    case STYLE:
      return "STYLE"; 
    case SUB:
      return "SUB"; 
    case SXDI:
      return "SXDI"; 
    case SXEXT:
      return "SXEXT"; 
    case SXIDSTM:
      return "SXIDSTM"; 
    case SXIVD:
      return "SXIVD"; 
    case SXLI:
      return "SXLI"; 
    case SXPI:
      return "SXPI"; 
    case SXSTRING:
      return "SXSTRING"; 
    case SXTBL:
      return "SXTBL"; 
    case SXTBPG:
      return "SXTBPG"; 
    case SXTBRGIITM:
      return "SXTBRGIITM"; 
    case SXVD:
      return "SXVD"; 
    case SXVI:
      return "SXVI"; 
    case SXVIEW:
      return "SXVIEW"; 
    case SXVS:
      return "SXVS"; 
    case TABID:
      return "TABID"; 
    case TABIDCONF:
      return "TABIDCONF"; 
    case TABLE:
      return "TABLE"; 
    case TEMPLATE:
      return "TEMPLATE"; 
    case TOPMARGIN:
      return "TOPMARGIN"; 
    case UDDESC:
      return "UDDESC"; 
    case UNCALCED:
      return "UNCALCED"; 
    case VCENTER:
      return "VCENTER"; 
    case VERTICALPAGEBREAKS:
      return "VERTICALPAGEBREAKS"; 
    case WINDOW1:
      return "WINDOW1"; 
    case WINDOW2:
      return "WINDOW2"; 
    case WINDOWPROTECT:
      return "WINDOWPROTECT"; 
    case WRITEACCESS:
      return "WRITEACCESS"; 
    case WRITEPROT:
      return "WRITEPROT"; 
    case WSBOOL:
      return "WSBOOL"; 
    case XCT:
      return "XCT"; 
    case XF:
      return "XF"; 
    default:
      return "UNKNOWN"; 
    }
}

struct rowdescr {
        int last, end;
	char **cells;
};	


int codepage=1251; /*default*/
struct rowdescr *rowptr=NULL;
char **saved_reference=NULL;
int startrow=0,lastrow=0;
void process_item (int rectype, int reclen, char *rec) {
    switch (rectype) {
        case CODEPAGE: {
	     int codepage;
	     codepage=getshort(rec,0);
	     fprintf(stderr,"Code page = %d\n",codepage);
	     break;
	}     
	case LABEL: {
	     int row,col,len;
	     struct rowdescr *prow;

	     saved_reference=NULL;
	     row = getshort(rec,0); 
	     col = getshort(rec,2);
	     len = getshort(rec,6);
	     allocate(row,col);
	     /* printf ("LABEL at %d %d length=%d place for it=%d\n",
	       row,col,len,reclen-8); */
	     prow=rowptr+row;
	     prow->cells[col]=malloc(len+1);
	     strncpy(prow->cells[col],rec+8,len);
	     (prow->cells[col])[len]=0;
	     break;
	}     
	case NUMBER: {
	     int row,col;
	     struct rowdescr *prow;

	     saved_reference=NULL;
	     row = getshort(rec,0)-startrow; 
	     col = getshort(rec,2);
	     allocate(row,col);
	     /* printf ("NUMBER at %d %d\n",row,col); */
	     prow=rowptr+row;
	     prow->cells[col]=strdup(format_double(rec,7));
	     break;
        }
	case FORMULA: { 
	     int row,col;
	     struct rowdescr *prow;

	     saved_reference=NULL;
	     row = getshort(rec,0)-startrow; 
	     col = getshort(rec,2);
	     allocate(row,col);
	     prow=rowptr+row;
             if (((unsigned char)rec[12]==0xFF)&&(unsigned char)rec[13]==0xFF) {
	        /* not a floating point value */
		if (rec[6]==1) {
		   /*boolean*/
                   char buf[2]="0";
		   buf[0]+=rec[9];
		   prow->cells[col]=strdup(buf);
		} else if (rec[6]==2) {
		   /*error*/
		   char buf[6]="ERROR";
		   prow->cells[col]=strdup(buf);
		} else if (rec[6]==0) {
		   saved_reference=prow->cells+col;
		}   
	     } else {
	         prow->cells[col]=strdup(format_double(rec,6));
             }		 
	     break;
       }
       case STRING: {
            int len;
	    if (!saved_reference) {
	       fprintf(stderr,"String record without preceeding string formula\n");
	       break;
	    }
	    len = getshort(rec,0);
	    *saved_reference=malloc(len+1);
	    strncpy(*saved_reference,rec+2,len);
	    (*saved_reference)[len]=0;
	    break;
       }	    
	    
       case BOF: {
          if (rowptr) {
	    fprintf(stderr,"BOF when current sheet is not flushed\n");
	    free_sheet();
	  }
	  break;
       }	  
       case MSEOF: {
           if (!rowptr) break;
	   print_sheet();
	   free_sheet();
	   break;
       }
       default: {
       }
   }
}  

void allocate (int row,int col) {
   int newrow,newcol;
   if (row>=lastrow) {
      newrow=(row/16+1)*16;
      rowptr=realloc(rowptr,newrow*sizeof(struct rowdescr));
      memset(rowptr+lastrow,0,(newrow-lastrow)*sizeof(struct rowdescr));
      lastrow=newrow;
   }
   if (col>=rowptr[row].end) {
      newcol=(col/16+1)*16;
      rowptr[row].cells=realloc(rowptr[row].cells,newcol*sizeof(char *));
      memset(rowptr[row].cells+rowptr[row].end,0,(newcol-rowptr[row].end)
	    *sizeof(char *));
      rowptr[row].end=newcol;
   }  
   if (col>rowptr[row].last) rowptr[row].last=col;
}
char *format_double(char *rec,int offset) {	
     static char buffer [128];
     union { char cc[8];
             double d;} dconv;
     char *d,*s;
     int i;
     /* FIXME - it is byteorder dependent */
     /* Intel version */
     for(s=rec+offset,d=dconv.cc,i=0;
       i<8;i++) *(d++)=*(s++);
     /* Sparc version */  
     /* for(s=rec+offset+8,d=dconv.cc,i=0;
       i<8;i++) *(d++)=*(--s);
      */ 
     sprintf(buffer,"%.10g",dconv.d);
     return buffer;
}

int getshort(char *rec,int offset) {
    return (signed short int) (*((unsigned char *)(rec+offset))|
              ((*((unsigned char *)(rec+offset+1)))<<8));
}	      

void free_sheet(void) {
   int i,j;
   struct rowdescr *row;
   char **col;
   for (row=rowptr,i=0;i<lastrow;i++,row++) {
       if (!row->cells) continue;
       for (col=row->cells,j=0;j<row->end;j++,col++) {
           if (*col) {
	      free(*col);
	   }
       }
       free (row->cells);
   }
   free (rowptr);
   rowptr=NULL;
}

void print_sheet(void) {
    int i,j;
    struct rowdescr *row;
    char **col;
    lastrow--;
    while (lastrow>0&&!rowptr[lastrow].cells) lastrow--;
    for(i=0,row=rowptr;i<=lastrow;i++,row++) {
       if (row->cells) {
	  for (j=0,col=row->cells;j<=row->last;j++,col++) {
	     if (j) fputc(',',stdout);
	     if (*col) {
	        print_value(*col);
	     }
	  }
       }
       fputc('\n',stdout);
    }
}
/* time_t float2date(double f) { */
/* Hacked version. Excell stores date as floating point count of days
 * since 1.1.1900. We are substracting value of 1.1.1970 and multiplying
 * by 86400 thus getting seconds from the epoch
 */
/* return rint((f-25569.0)*86400); 
}*/
void print_value(char *value) 
{
  int i,len;
  int quotes=0;
  len=strlen(value);
  for (i=0;i<len;i++) {
     if (!isdigit(value[i])&&toupper(value[i])!='E'&&
         value[i]!='+'&&value[i]!='-' && value[i]!='.') {
	quotes=1;
	break;
     }
  }
  if (quotes) {
    fputc('\"',stdout);
    for (i=0;i<len;i++) {
      if (value[i]=='\"') {
         fputc('\"',stdout);
         fputc('\"',stdout);
      } else {
         fputc(value[i],stdout);
      }
    }   
    fputc('\"',stdout);
  } else {
    fputs(value,stdout);
  }
}    

